#  Copyright (c) 2022. Edson Moreira
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
#  documentation files (the "Software"), to deal in the Software without restriction, including without limitation
#  the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and
#  to permit persons to whom the Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all copies or substantial portions of
#  the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
#  BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
#  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import sys
from pathlib import Path
from typing import Tuple

import click
import numpy as np
from PIL import Image

from photonbend.core.lens_image_type import LensImageType
from photonbend.core.sphere_image import SphereImage
from photonbend.projections.equirectangular import make_projection, compute_best_width
from photonbend.lens import equisolid, rectilinear, equidistant, \
    orthographic, stereographic
from photonbend.utils import degrees_to_radians
from .shared import lens_choices, type_choices, type_choices_help, double_type_fov_warning, rotation_help


def _check_fov(fov: float, image_type: LensImageType):
    if image_type is LensImageType.DOUBLE_INSCRIBED and fov < 180:
        raise ValueError("The fov of a double image can't be smaller than 180 degrees.")
    if fov > 360:
        raise ValueError("The fov of an image can't be higher than 360 degrees.")
    r_fov = degrees_to_radians(fov)
    return r_fov


def check_output(output: Path):
    out = Path(output)
    if not (out.suffix.lower() in ['.jpg', '.jpeg', '.png']):
        print("The desired output image should be a JPG or PNG file.")
        print("Provide an output filename ending in either JPG, JPEG or PNG (case insensitive)")
        print("Exiting!")
        sys.exit(1)
    if out.exists():
        while True:
            ans = input("File already exists. Overwrite? (y/n) ")
            if ans in ['y', 'n']:
                break
        if ans == 'n':
            print('Exiting!')
            sys.exit(0)
    return out


@click.argument('input_image', type=click.Path(exists=True))
@click.option('--type', required=True, help='The type of the input image. ' + type_choices_help,
              type=type_choices)
@click.option('--lens', required=True, help='The lens type that was used on the input photo.',
              type=lens_choices)
@click.option('--fov', required=True, type=click.FLOAT,
              help='The lens field of view of the input photo in degrees. ' + double_type_fov_warning)
@click.option('--ssample', required=False, type=click.INT, help='The ammount of supersampling applied', default=1)
@click.option('-r', '--rotation', required=False, type=click.FLOAT, nargs=3, default=(0, 0, 0), help=rotation_help)
@click.argument('output_image', type=click.Path(exists=False))
def make_pano(input_image: click.Path, type: str, lens: str, fov: float, output_image: click.Path,
              ssample: int, rotation: Tuple[float, float, float]) -> None:
    """Make a panorama out of a photo.

    \b
    INPUT is the path to the source photo.
    OUTPUT is the desired path of the destiny panorama.
    """
    out = check_output(output_image)

    types_dict = {'inscribed': LensImageType.INSCRIBED,
                  'double': LensImageType.DOUBLE_INSCRIBED,
                  'cropped': LensImageType.CROPPED_CIRCLE,
                  'full': LensImageType.FULL_FRAME}

    lens_types = {'equidistant': equidistant,
                  'equisolid': equisolid,
                  'orthographic': orthographic,
                  'rectilinear': rectilinear,
                  'stereographic': stereographic}

    source_lens = lens_types[lens]
    source_type = types_dict[type]
    source_fov = _check_fov(fov, source_type)

    try:
        with Image.open(input_image) as image:
            source_array = np.asarray(image)
    except IOError:
        print('Error: Input image could not be opened!')
        print('Exiting!')
        sys.exit(1)

    source_sphere = SphereImage(source_array, source_type, source_fov, source_lens)

    if rotation != (0, 0, 0):
        rotation_rad = list(map(degrees_to_radians, rotation))
        pitch, yaw, roll = rotation_rad
        source_sphere.set_rotation(pitch, roll, yaw)

    width = compute_best_width(source_sphere)

    destiny_arr = make_projection(source_sphere, 0, width)
    destiny_image = Image.fromarray(destiny_arr)
    print('Finished!')

    try:
        destiny_image.save(output_image)
    except IOError:
        print('Could not save to the specified location!')
        print('Exiting!')
        print('Exiting!')
        sys.exit(1)



#  Copyright (c) 2022. Edson Moreira
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
#  documentation files (the "Software"), to deal in the Software without restriction, including without limitation
#  the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and
#  to permit persons to whom the Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all copies or substantial portions of
#  the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
#  BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
#  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


# TODO Improve on the use of exceptions
from typing import Tuple

import numpy as np

from numba import uint8, float64, typeof, complex128, cfunc, int64, bool_
from numba.experimental import jitclass

from photonbend.utils import vector_magnitude, decompose, weighted_sum
from photonbend.core.lens_image_type import LensImageType
from photonbend.utils.utils import _2ints

DoubleCardinal = Tuple[Tuple[int, int], Tuple[int, int]]
FULL_CIRCLE = (np.pi * 2)
INVALID_POSITION = (-1, -1)


@cfunc(float64(float64, bool_))
def _a_numba_function_definition(_b, _c):
    """A SIMPLE IDENTITY FUNCTION THAT IS NOT MEANT TO BE USED!

    It is present only to make it easier to categorize functions marked with its decorator
    <@cfunc(float64(float64, bool_))> so we can have those as class members on a numba jitclass
    """
    return _b


spec = [
    ('north_pole', complex128),
    ('south_pole', complex128),
    ('fov', float64),
    ('dpf', float64),
    ('image', uint8[:, :, :]),
    ('image_type', int64),
    ('lens', typeof(_a_numba_function_definition)),
]


@jitclass(spec)
class LensImage:
    """A class that maps images to a sphere
    This class allows us to get or set pixel values that are mapped to specific coordinates of that sphere. This
    enables us to convert images between different kinds of lenses or to create projections from it.
    """

    def __init__(self, image_arr, i_type, fov, lens):
        if i_type == LensImageType.DOUBLE_INSCRIBED:
            if fov < np.pi:
                raise ValueError("The FOV of a DOUBLE_INSCRIBED image should be of at minimum pi radians")

        self.image = image_arr
        self.image_type = i_type
        self.lens = lens
        self.fov = fov
        self._set_poles()
        self._set_dpf()

    def _set_poles(self):
        height, width = self.image.shape[:2]

        if self.image_type == LensImageType.DOUBLE_INSCRIBED:  # double image
            if width > height:  # horizontal
                real_width = width / 2
                self.north_pole = complex(real_width, height) / 2
                self.north_pole -= complex(0.5, 0.5)
                self.south_pole = self.north_pole + real_width
            else:
                raise ValueError("The LensImage class doesn't support vertical double inscribed images")
        else:  # Simple image
            self.north_pole = complex(width, height) / 2
            self.north_pole -= complex(0.5, 0.5)
            self.south_pole = complex(0, 0)

    def _set_dpf(self) -> None:
        """
        This function sets the dpf (dots per focal distance)

        Only the self.init method should call this.

        It computes the maximum distance the maximum angle this lens is set to produce in focal distances.
        To simplify the calculations, we always use a focal distance of one, and make the dots per focal distance (dpf)
        variable.
        So, in order to calculate the dpf, we measure the maximum distance the lens produce if focal distances,
        we calculate the longest vector of this image (from the center of the image to one of it's sides) and we
        divide the second by the first to arrive at the dpf.  Then we set this to the current object.

        :return: None
        """
        maximum_lens_angle = self.fov / 2

        maximum_image_magnitude = self.maximum_magnitude
        lens_max_angle_magnitude = self.lens(maximum_lens_angle, False)
        self.dpf = maximum_image_magnitude / lens_max_angle_magnitude

    @property
    def maximum_magnitude(self):
        half = 0.5
        half_c = complex(half, half)
        if self.image_type == LensImageType.FULL_FRAME:
            return vector_magnitude(self.north_pole + half_c)
        elif self.image_type == LensImageType.CROPPED_CIRCLE:
            return self.north_pole.imag + half
        elif self.image_type == LensImageType.INSCRIBED:
            return self.north_pole.imag + half
        elif self.image_type == LensImageType.DOUBLE_INSCRIBED:
            return self.north_pole.imag + half
        return self.north_pole.imag + half

    @property
    def shape(self):
        return self.image.shape

    def get_image_array(self):
        """ Returns a copy of the underlying image matrix
        :return: A copy of the image array
        """
        return np.copy(self.image)

    def is_position(self, x, y):
        """
        Checks whether the passed parameters are within the limits of the underlying image
        :param x: absolute x position of the image you want to check
        :param y: absolute y position of the image you want to check
        :return: True if withing the image, False otherwise
        """
        height, width = self.image.shape[:2]
        if (0 > x or x >= width) or (0 > y or y >= height):
            return False
        return True

    def is_valid_position(self, x, y) -> bool:
        """
        Checks whether the passed parameters are within the limits of the underlying image and the lens specification
        :param x: absolute x position of the image you want to check
        :param y: absolute y position of the image you want to check
        :return: True if withing the image, False otherwise
        """
        if not self.is_position(x, y):
            return False

        relative_position, _ = self.absolute_to_relative(complex(x, y))
        if round(vector_magnitude(relative_position)) <= self.maximum_magnitude:
            return False

        return True

    def get_from_cartesian(self, x: int, y: int) -> uint8[:]:
        """ Get the value represented on the cartesian position x, y of this image

        :param x: The horizontal position of the desired value
        :param y: The vertical position of the desired value
        :return: a triplet of values of type uint8
        """
        if self.is_position(x, y):
            return self.image[y, x, :]
        else:
            raise Exception("The requested cartesian coordinates are outside the bounds of the image!")

    def set_to_cartesian(self, x: int, y: int, value: uint8[:]) -> None:
        """ Get the value represented on the cartesian position x, y of this image

        :param value:
        :param x: The horizontal position of the desired value
        :param y: The vertical position of the desired value
        :return: a triplet of values of type uint8
        """
        if self.is_position(x, y):
            self.image[y, x, :] = value
        else:
            raise Exception("The target cartesian coordinates are outside the bounds of the image!")

    def get_from_spherical(self, latitude: float, longitude: float) -> uint8[:]:
        pos1, pos2 = self.translate_spherical_to_cartesian(latitude, longitude)
        try:
            if self.image_type != LensImageType.DOUBLE_INSCRIBED:
                return self.get_from_cartesian(*_2ints(*pos1))
            else:  # DOUBLE_INSCRIBED
                r_value = self._ds_get_from_spherical(latitude, pos1, pos2)
                return r_value
        except:
            r = np.zeros(3, np.core.uint8)
            r[:] = 0, 0, 0
            return r

    def _ds_get_from_spherical(self, latitude, pos1, pos2):
        r_value = np.zeros(3, np.core.uint8)
        double_image_latitude = (self.fov - np.pi) / 2
        if (latitude > double_image_latitude) or (latitude < (-double_image_latitude)):
            if latitude >= 0:
                r_value = self.get_from_cartesian(*_2ints(*pos1))
            else:
                r_value = self.get_from_cartesian(*_2ints(*pos2))
        else:
            cross_range = 2 * double_image_latitude
            factor_positive = cross_range - (double_image_latitude - latitude)
            factor_negative = cross_range + (-double_image_latitude - latitude)

            v_pos = self.get_from_cartesian(*_2ints(*pos1))
            v_neg = self.get_from_cartesian(*_2ints(*pos2))

            f_value = weighted_sum(v_pos, v_neg, factor_positive, factor_negative)
            r_value = f_value
        return r_value

    def set_to_spherical(self, latitude, longitude, data):
        """Sets data to the position in the image that is represented by latitude and longitude.
        If the given latitude and longitude points are not represented in the image, it raises an exception.

        :param latitude: The target latitude
        :param longitude: The target longitude
        :param data: The data that is to be put
        :return: None
        """
        try:
            if not self.image_type == LensImageType.DOUBLE_INSCRIBED:
                pos1, _ = self.translate_spherical_to_cartesian(latitude, longitude)
                if pos1 != INVALID_POSITION:
                    self.set_to_cartesian(*_2ints(*pos1), data)
            else:  # DOUBLE_INSCRIBED
                pos1, pos2 = self.translate_spherical_to_cartesian(latitude, longitude)
                if pos1 != INVALID_POSITION:
                    self.set_to_cartesian(*_2ints(*pos1), data)
                if pos2 != INVALID_POSITION:
                    self.set_to_cartesian(*_2ints(*pos2), data)
        except Exception:
            raise Exception("The given spherical coordinates are not represented on the cartesian image")

    def translate_spherical_to_cartesian(self, latitude: float, longitude: float) -> DoubleCardinal:
        """
        :param latitude:
        :param longitude:
        :return:
        """

        if self.image_type != LensImageType.DOUBLE_INSCRIBED:
            polar_distance = self.lens(np.pi / 2 - latitude, False) * self.dpf
            factors = np.exp(longitude * 1j)
            relative_position = factors * polar_distance
            position = self.relative_to_absolute(relative_position, self.north_pole)
            x, y = decompose(position)
            return (x, y), INVALID_POSITION

        else:  # DOUBLE_INSCRIBED
            pos1, pos2 = self._ds_translate_spherical_to_cartesian(latitude, longitude)
            return pos1, pos2

    def _ds_translate_spherical_to_cartesian(self, latitude, longitude):
        double_image_latitude = self.fov - np.pi
        # latitude > 0:
        if latitude > (0 - double_image_latitude):
            polar_distance = self.lens(np.pi / 2 - latitude, False) * self.dpf
            factors = np.exp(longitude * 1j)
            relative_position = factors * polar_distance
            position = self.relative_to_absolute(relative_position, self.north_pole)
            pos1 = decompose(position)
        else:
            pos1 = INVALID_POSITION
        # latitude < 0
        if latitude < (0 + double_image_latitude):
            polar_distance = self.lens(np.pi / 2 + latitude, False) * self.dpf
            factors = np.exp(longitude * 1j)
            factors = complex(-factors.real, factors.imag)  # double_inscribed X inversion
            relative_position = factors * polar_distance
            position = self.relative_to_absolute(relative_position, self.south_pole)
            pos2 = decompose(position)
        else:
            pos2 = INVALID_POSITION
        return pos1, pos2

    def translate_cartesian_to_spherical(self, x: float, y: float) -> Tuple[float, float]:
        max_latitude = np.pi / 2
        min_latitude = max_latitude - self.fov / 2
        if self.image_type == LensImageType.DOUBLE_INSCRIBED:
            min_latitude = -max_latitude

        absolute_position = complex(x, y)
        relative_position, reference_point = self.absolute_to_relative(absolute_position)

        magnitude = vector_magnitude(relative_position)
        if magnitude > self.maximum_magnitude:
            raise Exception('not a valid position')

        if reference_point == self.north_pole:
            latitude = max_latitude - self.lens(magnitude / self.dpf, True)
        else:  # reference_point == self.south_pole
            latitude = min_latitude + self.lens(magnitude / self.dpf, True)

        if self.image_type == LensImageType.DOUBLE_INSCRIBED and reference_point == self.south_pole:
            relative_position = complex(-relative_position.real, relative_position.imag)  # double inscribed inversion
        euler_relative_position = relative_position / magnitude
        longitude = np.log(euler_relative_position).imag

        return latitude, longitude

    def relative_to_absolute(self, relative_position: complex, reference_point: complex) -> complex:
        if reference_point == self.north_pole or reference_point == self.south_pole:
            return (relative_position.real + reference_point.real) + 1j * (
                    reference_point.imag - relative_position.imag)
        raise Exception("Only the image's north and south poles may be used as a reference point")

    def absolute_to_relative(self, absolute_position: complex) -> Tuple[complex, complex]:
        x = absolute_position.real

        reference_point = self.north_pole
        if self.image_type == LensImageType.DOUBLE_INSCRIBED:
            halfway_x = self.image.shape[1] / 2
            if x >= halfway_x:
                reference_point = self.south_pole

        return ((absolute_position.real - reference_point.real) + 1j * (reference_point.imag - absolute_position.imag),
                reference_point)



#  Copyright (c) 2022. Edson Moreira
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
#  documentation files (the "Software"), to deal in the Software without restriction, including without limitation
#  the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and
#  to permit persons to whom the Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all copies or substantial portions of
#  the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
#  BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
#  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
import click

lens_choices = click.Choice(['equidistant', 'equisolid', 'orthographic', 'rectilinear', 'stereographic'])
type_choices = click.Choice(['inscribed', 'double', 'cropped', 'full'])

type_choices_help = \
    """

    \b
    The choices are:
    - inscribed: The valid data is on a inscribed circle inscribed.
    - double: The valid data is on two inscribed side-by-side circles.
    - cropped: The valid data is on a inscribed circle, top-and-bottom cropped. 
    - full: The whole area of the image is valid data. 
    """

double_type_fov_warning = \
    """

    IMPORTANT: When dealing with double images, you should pass the FOV of a single circle and it must be larger than 180."""

rotation_help = \
    """
    The rotation that should be applied to the camera.
    This is a 3-valued parameter in the form <pitch yaw roll>
    """



#  Copyright (c) 2022. Edson Moreira
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
#  documentation files (the "Software"), to deal in the Software without restriction, including without limitation
#  the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and
#  to permit persons to whom the Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all copies or substantial portions of
#  the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
#  BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
#  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

#
#  Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
#  documentation files (the "Software"), to deal in the Software without restriction, including without limitation
#  the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and
#  to permit persons to whom the Software is furnished to do so, subject to the following conditions:
#
#
import sys
from pathlib import Path
from typing import Tuple

import click
import numpy as np
from PIL import Image

from photonbend.core.lens_image_type import LensImageType
from photonbend.core.sphere_image import SphereImage
from photonbend.lens import equisolid, rectilinear, equidistant, \
    orthographic, stereographic
from photonbend.utils import degrees_to_radians
from .shared import lens_choices, type_choices, type_choices_help, double_type_fov_warning, rotation_help


def _check_fov(fov: float, image_type: LensImageType):
    if image_type is LensImageType.DOUBLE_INSCRIBED and fov < 180:
        raise ValueError("The fov of a double image can't be smaller than 180 degrees.")
    if fov > 360:
        raise ValueError("The fov of an image can't be higher than 360 degrees.")
    r_fov = degrees_to_radians(fov)
    return r_fov


def check_output(output: Path):
    out = Path(output)
    if not (out.suffix.lower() in ['.jpg', '.jpeg', '.png']):
        print("The desired output image should be a JPG or PNG file.")
        print("Provide an output filename ending in either JPG, JPEG or PNG (case insensitive)")
        print("Exiting!")
        sys.exit(1)
    if out.exists():
        while True:
            ans = input("File already exists. Overwrite? (y/n) ")
            if ans in ['y', 'n']:
                break
        if ans == 'n':
            print('Exiting!')
            sys.exit(0)
    return out


@click.argument('input_image', type=click.Path(exists=True))
@click.option('--itype', required=True, help='The type of the input image. ' + type_choices_help,
              type=type_choices)
@click.option('--ilens', required=True, help='The lens type that was used on the input photo.',
              type=lens_choices)
@click.option('--ifov', required=True, type=click.FLOAT,
              help='The lens field of view of the input photo in degrees. ' + double_type_fov_warning)
@click.option('--otype', required=True, help='The type of the output image.' + type_choices_help,
              type=type_choices)
@click.option('--olens', required=True,
              help='The lens type that was used on the input photo. ' + double_type_fov_warning,
              type=lens_choices)
@click.option('--ofov', required=True, type=click.FLOAT, help='The lens field of view of the output photo in degrees.')
@click.option('--ssample', required=False, type=click.INT, help='The ammount of supersampling applied (ss²)', default=1)
@click.argument('output_image', type=click.Path(exists=False))
@click.option('-r', '--rotation', required=False, type=click.FLOAT, nargs=3, default=(0, 0, 0), help=rotation_help)
def alter_photo(input_image: click.Path, itype: str, ilens: str, ifov: float, otype: str, olens: str, ofov: float,
                output_image: click.Path, ssample: int, rotation: Tuple[float, float, float]) -> None:
    """Change the the lens and FoV of a photo.

    \b
    INPUT is the path to the source photo.
    OUTPUT is the desired path of the destiny photo.
    """
    out = check_output(output_image)

    types_dict = {'inscribed': LensImageType.INSCRIBED,
                  'double': LensImageType.DOUBLE_INSCRIBED,
                  'cropped': LensImageType.CROPPED_CIRCLE,
                  'full': LensImageType.FULL_FRAME}

    lens_types = {'equidistant': equidistant,
                  'equisolid': equisolid,
                  'orthographic': orthographic,
                  'rectilinear': rectilinear,
                  'stereographic': stereographic}

    source_lens = lens_types[ilens]
    source_type = types_dict[itype]
    source_fov = _check_fov(ifov, source_type)

    destiny_lens = lens_types[olens]
    destiny_type = types_dict[itype]
    destiny_fov = _check_fov(ofov, destiny_type)

    try:
        with Image.open(input_image) as image:
            source_array = np.asarray(image)
    except IOError:
        print('Error: Input image could not be opened!')
        print('Exiting!')
        sys.exit(1)

    source_sphere = SphereImage(source_array, source_type, source_fov, source_lens)
    destiny_sphere = SphereImage(np.zeros(source_array.shape, np.core.uint8), destiny_type, destiny_fov, destiny_lens)

    if rotation != (0, 0, 0):
        rotation_rad = list(map(degrees_to_radians, rotation))
        pitch, yaw, roll = rotation_rad
        source_sphere.set_rotation(pitch, roll, yaw)

    destiny_sphere.map_from_sphere_image(source_sphere, ssample)
    destiny_arr = destiny_sphere.get_image_array()
    destiny_image = Image.fromarray(destiny_arr)

    try:
        destiny_image.save(output_image)
    except IOError:
        print('Could not save to the specified location!')
        print('Exiting!')
        sys.exit(1)



#  Copyright (c) 2022. Edson Moreira
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
#  documentation files (the "Software"), to deal in the Software without restriction, including without limitation
#   the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and
#   to permit persons to whom the Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all copies or substantial portions of
#  the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
#  BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
#  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


# A lot of the code in here doesn't make the use of the standards and best tools available in the Python language.
# Since we are dealing with a lot of arrays and calculations, it was thought that Numba (https://numba.pydata.org/)
# should be used so that the code can run in reasonable amount of time. This goal has benn achieved here at the cost of
# some python functionality.

# TODO Improve on the use of exceptions
from typing import Tuple

import numpy as np

from numba import uint8, float64, njit, prange, complex128, bool_, typeof
from numba.experimental import jitclass

from photonbend.core import LensImage

DoubleCardinal = Tuple[Tuple[int, int], Tuple[int, int]]

FULL_CIRCLE = (np.pi * 2)


@njit
def _calculate_rotation_matrix(pitch: float, yaw: float, roll: float):
    """Computes a rotation matrix from the three primary rotation axis
    For more info see:  https://en.wikipedia.org/wiki/Rotation_matrix
    or: https://mathworld.wolfram.com/RotationMatrix.html

    :param pitch: represents a rotation in the x axis in radians
    :param yaw: represents a rotation in the y axis in radians
    :param roll: represents a rotation in the z axis in radians
    :return: a rotation matrix
    """

    # TODO unwind the matrices and wind them up with array methods so they are contiguous
    cos_pitch = np.cos(pitch)
    sin_pitch = np.sin(pitch)
    pitch_matrix = np.array((1, 0, 0, 0, cos_pitch, sin_pitch, 0, -sin_pitch, cos_pitch)).reshape((3, 3))

    cos_yaw = np.cos(yaw)
    sin_yaw = np.sin(yaw)
    yaw_matrix = np.array((cos_yaw, 0, -sin_yaw, 0, 1, 0, sin_yaw, 0, cos_yaw)).reshape((3, 3))

    cos_roll = np.cos(roll)
    sin_roll = np.sin(roll)
    roll_matrix = np.array((cos_roll, sin_roll, 0, -sin_roll, cos_roll, 0, 0, 0, 1)).reshape((3, 3))

    rotation_matrix = pitch_matrix @ yaw_matrix @ roll_matrix

    return rotation_matrix


# Needed for using numba jitclass
sphere_image_spec = [
    ('north_pole', complex128),
    ('south_pole', complex128),
    ('fov', float64),
    ('dpf', float64),
    ('rotated', bool_),
    ('rotation_matrix', float64[:, :]),
    ('i_rotation_matrix', float64[:, :]),
]


@jitclass(sphere_image_spec)
class SphereImage:
    lens_image: LensImage
    """A class adds rotation functionality to the LensImage class.
    This class acts like a decorator class to the lens image class enabling us to rotate the angles of the underlying
    image. This allows us to get different angles, and convert between different kinds of lenses and image formats 
    which is essential when dealing with 360 degree images or can greatly simplify creating projections. 
    """

    def __init__(self, image_arr, image_type, fov, lens):
        """Create a new SphereImage

        :param image_arr: The array representing the image.
        :param image_type: An enum of <photonbend.core.ImageType>
        :param fov: The maximum Field of View of this image
        :param lens: The lens function of this sphere image
        """
        self.lens_image = LensImage(image_arr, image_type, fov, lens)
        self.rotated = False
        self.rotation_matrix = np.zeros((3, 3), np.core.float64)
        self.i_rotation_matrix = np.zeros((3, 3), np.core.float64)
        self.set_rotation(0.0, 0.0, 0.0)

    def _get_rotated_spherical_coordinates(self, latitude: float, longitude: float, inverse: bool = False):
        """ Translate coordinates using the instance rotation matrix

        :param latitude: The original latitude in radians as a float
        :param longitude: The original longitude in radians as a float
        :param inverse: Flag that sets the inverse transformation. This is used when converting coordinates generated
                        from the underlying image quasi-polar coordinates to sphere coordinates.
        :return: translated coordinates in radians as a 2-tuple of floats
        """
        # TODO refactor to make use of complex numbers
        if not self.rotated:
            return latitude, longitude

        y = np.sin(latitude)
        xz = np.exp(longitude * 1j)
        x = xz.real * np.cos(latitude)
        z = xz.imag * np.cos(latitude)

        position_vector = np.zeros(3, np.core.float64)
        position_vector[:] = x, y, z
        if not inverse:
            new_position_vector = self.rotation_matrix.dot(position_vector)
        else:
            new_position_vector = self.i_rotation_matrix.dot(position_vector)

        translated_latitude = np.arcsin(new_position_vector[1])
        translated_xz_magnitude = np.cos(translated_latitude)
        translated_xz = complex(new_position_vector[0] / translated_xz_magnitude,
                                new_position_vector[2] / translated_xz_magnitude)
        translated_longitude = np.log(translated_xz).imag

        return translated_latitude, translated_longitude

    def set_rotation(self, pitch: float, yaw: float, roll: float) -> None:
        """
        Sets the instance rotation matrix to the pitch, yaw, roll values passed

        If used with all values as 0 (zero), it resets the sphere to it's initial rotation.
        It also sets the inverse rotation matrix.
        :param pitch: The amount of rotation in radians along the x axis as a float
        :param yaw: The amount of rotation in radians along the y axis as a float
        :param roll: The amount of rotation in radians along the z axis as a float
        :return: None
        """
        if pitch == yaw == roll == 0.0:
            self.rotated = False
        else:
            self.rotated = True

        rot = _calculate_rotation_matrix(pitch, yaw, roll)
        i_rot = _calculate_rotation_matrix(-pitch, -yaw, -roll)
        self.rotation_matrix = rot
        self.i_rotation_matrix = i_rot

    def add_rotation(self, pitch, yaw, roll):
        # TODO change method to chain rotations instead of simply adding them together.
        """Add a new rotation to the instance's current rotation matrix.
        With this method a sphere can be rotated arbitrarily many times along its axis.
        This method does not reset the rotation. It actually updates it, adding this new rotation to the previously
        existing one.
        The problem is it won't add consecutively.

        If used with all values as 0 (zero), it doesn't do anything
        It also updates the inverse rotation matrix.
        :param pitch: The amount of rotation in radians along the x axis as a float
        :param yaw: The amount of rotation in radians along the y axis as a float
        :param roll: The amount of rotation in radians along the z axis as a float
        :return: None
        """

        self.rotated = True
        rot = _calculate_rotation_matrix(pitch, yaw, roll)
        i_rot = _calculate_rotation_matrix(-pitch, -yaw, -roll)
        self.rotation_matrix[:] = self.rotation_matrix @ rot
        self.i_rotation_matrix[:] = self.i_rotation_matrix @ i_rot

    def get_image_array(self):
        """ Returns a copy of the underlying image matrix
        :return: A copy of the image array
        """
        return self.lens_image.get_image_array()

    @property
    def shape(self):
        # returns the underlying image shape
        return self.lens_image.shape

    def check_position(self, x, y):
        """
        Checks whether the passed parameters are withing the limits of the underlying image
        :param x: absolute x position of the image you want to check
        :param y: absolute y position of the image you want to check
        :return: True if withing the image, False otherwise
        """
        height, width = self.lens_image.shape[:2]
        if (0 > x or x >= width) or (0 > y or y >= height):
            return False
        return True

    def get_from_cartesian(self, x: int, y: int) -> uint8[:]:
        """ Get the value represented on the cartesian position x, y of this image

        :param x: The horizontal position of the desired value
        :param y: The vertical position of the desired value
        :return: a triplet of values of type uint8
        """
        return self.lens_image.get_from_cartesian(x, y)

    def set_to_cartesian(self, x: int, y: int, value: uint8[:]) -> None:
        """ Get the value represented on the cartesian position x, y of this image

        :param value:
        :param x: The horizontal position of the desired value
        :param y: The vertical position of the desired value
        :return: a triplet of values of type uint8
        """
        self.lens_image.set_to_cartesian(x, y, value)

    def get_from_spherical(self, latitude: float, longitude: float) -> uint8[:]:
        """Returns the value at the given latitude and longitude.

        If the request point is invalid, it returns the triplet (0, 0 ,0)
        :param latitude: The latitude of the requested value
        :param longitude: The longitude of the requested value
        :return: a triplet of values of type uint8
        """

        r_latitude, r_longitude = self._get_rotated_spherical_coordinates(latitude, longitude, True)
        return self.lens_image.get_from_spherical(r_latitude, r_longitude)

    def set_to_spherical(self, latitude: float, longitude: float, data: uint8[:]) -> None:
        """Sets the value at the given latitude and longitude to the value present on data

        :param latitude: The latitude of the value to set
        :param longitude: The longitude of the value to set
        :param data: The value that should be set
        :return: None
        """
        r_latitude, r_longitude = self._get_rotated_spherical_coordinates(latitude, longitude, True)
        self.lens_image.set_to_spherical(r_latitude, r_longitude, data)

    def translate_spherical_to_cartesian(self, latitude: float, longitude: float) -> DoubleCardinal:
        """Translate spherical coordinates to cartesian coordinates of the image

        :param latitude: The latitude of the location to be translated
        :param longitude: The longitude of the location to be translated
        :return: The cartesian coordinates of the image represented by the spherical coordinates passed
        """
        if not (-np.pi / 2) <= latitude <= (np.pi / 2):
            raise ValueError("latitude should be between pi/2 and -pi/2")

        r_latitude, r_longitude = self._get_rotated_spherical_coordinates(latitude, longitude, True)
        xy1, xy2 = self.lens_image.translate_spherical_to_cartesian(r_latitude, r_longitude)

        return xy1, xy2

    def translate_cartesian_to_spherical(self, x: float, y: float) -> Tuple[float, float]:
        """Translate cartesian coordinates to spherical coordinates of the image

        :param x: The horizontal coordinate
        :param y: The vertical coordinate
        :return: The spherical coordinates representing that location
        """
        latitude, longitude = self.lens_image.translate_cartesian_to_spherical(x, y)
        r_latitude, r_longitude = self._get_rotated_spherical_coordinates(latitude, longitude)

        return r_latitude, r_longitude

    def map_from_sphere_image(self, target_image, super_sampler: int) -> None:
        """ Maps one spherical image to another, enabling conversion between different lenses, FoVs and target.

        The current sphere image will receive the data from the target sphere image. However it won't just copy it.
        The information will be translated to this sphere using their angular data and other specs, like its lens and
        FoV, producing a new image.

        :param target_image: The sphere image that will serve as the source of the information.
        :param super_sampler: The super sampling factor, which enables improving the image at the cost of processing.
            WARNING: The super sampling factor is actually the square of this parameter.
            e.g.: When using a factor of 3, it will actually compute 9 times more. For 4 it is 16 and 5 makes it 25.
            The super sampler is implemented in a way that uses little memory space, so it doesn't hinder it's use
            on low-spec machines. It won't increase much the memory footprint.
        :return: None
        """
        _helper_map_from_sphere_image(self, target_image, super_sampler)


@njit(parallel=True)
def _helper_map_from_sphere_image(this_image: SphereImage, that_image: SphereImage, super_sampler: int):
    height, width = this_image.lens_image.shape[:2]

    # SUPER SAMPLING DATA
    ss_matrix_center_pixel = complex((super_sampler - 1) / 2, (super_sampler - 1) / 2)
    ss_subpixel_distance = 1 / super_sampler

    for column in prange(width):
        for row in prange(height):
            if super_sampler == 1:
                # No super sampling case
                try:
                    lat, lon = this_image.translate_cartesian_to_spherical(column, row)
                except Exception:  # Because of Numba's njit Exception limitation, that's all we currently use
                    continue
                this_image.set_to_cartesian(column, row, that_image.get_from_spherical(lat, lon))

            else:
                # super sampling
                super_sample_matrix = np.zeros((super_sampler, super_sampler, 3), np.core.uint8)
                for ss_row in prange(super_sampler):
                    for ss_column in prange(super_sampler):
                        ss_position = (complex(ss_column, ss_row) - ss_matrix_center_pixel) * ss_subpixel_distance

                        position_destiny = complex(column, row) + ss_position
                        a_column = position_destiny.real
                        a_row = position_destiny.imag
                        try:
                            lat, lon = this_image.translate_cartesian_to_spherical(a_column, a_row)
                        except Exception:  # Because of Numba's njit Exception limitation, that's all we currently use
                            continue

                        super_sample_matrix[ss_row, ss_column, :] = that_image.get_from_spherical(lat, lon)

                # Calculate the pixel mean and assign it!
                ss_mean = np.zeros(3, np.core.uint8)
                ss_mean[0] = int(round(np.mean(super_sample_matrix[:, :, 0])))
                ss_mean[1] = int(round(np.mean(super_sample_matrix[:, :, 1])))
                ss_mean[2] = int(round(np.mean(super_sample_matrix[:, :, 2])))
                this_image.set_to_cartesian(column, row, ss_mean[:])
    return



#  Copyright (c) 2022. Edson Moreira
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
#  documentation files (the "Software"), to deal in the Software without restriction, including without limitation
#  the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and
#  to permit persons to whom the Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all copies or substantial portions of
#  the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
#  BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
#  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import click
import logging
from .commands.alter_photo import alter_photo
from .commands.make_pano import make_pano
from .commands.make_photo import make_photo
import os


@click.group()
def main():
    """The photonbend utility allows one to change photos and panoramas.
    It provides multiple commands to change different kinds of images.
    """
    os.environ['NUMBA_DISABLE_PERFORMANCE_WARNINGS'] = '1'


main.command()(make_pano)
main.command()(alter_photo)
main.command()(make_photo)



#  Copyright (c) 2022. Edson Moreira
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
#  documentation files (the "Software"), to deal in the Software without restriction, including without limitation
#  the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and
#  to permit persons to whom the Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all copies or substantial portions
#  of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
#  BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
#  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import numpy as np
from numba import njit, float64, cfunc, bool_
from photonbend.utils import degrees_to_radians


@cfunc(float64(float64))
def _rectilinear_inverse(projection_in_focal_distance_units):
    theta = np.arctan(projection_in_focal_distance_units)
    return theta


@cfunc(float64(float64))
def _rectilinear(theta):
    """Mapping that uses the angle tangent

    As it uses the angle tangent, it should not be used with lens angles closing on 180 degrees.
    It is best to limit its use to lens angles of at most 165 degrees.
    As the angle presented is halved, it should not see a theta angle larger than 82.5 degrees.
    """
    if theta < 0:
        raise ValueError('The angle theta cannot be negative')
    if theta > degrees_to_radians(89):
        raise ValueError('The rectilinear function was not made to handle FOV larger than 179 degrees')
    return np.tan(theta)


@cfunc(float64(float64))
def _stereographic_inverse(projection_in_focal_distance_units):
    half_tan_theta = projection_in_focal_distance_units / 2
    half_theta = np.arctan(half_tan_theta)
    theta = 2 * half_theta
    return theta


@cfunc(float64(float64))
def _stereographic(theta):
    half_theta = theta / 2
    half_projection = np.tan(half_theta)
    projection = 2 * half_projection
    return projection


@cfunc(float64(float64))
def _equidistant_inverse(projection_in_focal_distance_units):
    return projection_in_focal_distance_units


@cfunc(float64(float64))
def _equidistant(theta):
    return theta


@cfunc(float64(float64))
def _equisolid_inverse(projection_in_focal_distance_units):
    half_sin_theta = projection_in_focal_distance_units / 2
    half_theta = np.arcsin(half_sin_theta)
    theta = 2 * half_theta
    return theta


@cfunc(float64(float64))
def _equisolid(theta):
    half_theta = theta / 2
    half_projection = np.sin(half_theta)
    projection = 2 * half_projection
    return projection


@cfunc(float64(float64))
def _orthographic_inverse(projection_in_focal_distance_units):
    theta = np.arcsin(projection_in_focal_distance_units)
    return theta


@cfunc(float64(float64))
def _orthographic(theta):
    projection = np.sin(theta)
    return projection


@cfunc(float64(float64))
def _thoby_inverse(projection_in_focal_distance_units):
    k1 = 1.47
    k2 = 0.713
    theta = np.arcsin(projection_in_focal_distance_units / k1) / k2

    return theta


@cfunc(float64(float64))
def _thoby(theta):
    k1 = 1.47
    k2 = 0.713
    projection = k1 * np.sin(k2 * theta)
    return projection


# Begin the exported functions

@cfunc(float64(float64, bool_))
def rectilinear(angle_or_projection, inverse=False):
    if not inverse:
        return _rectilinear(angle_or_projection)
    return _rectilinear_inverse(angle_or_projection)


@cfunc(float64(float64, bool_))
def equisolid(angle_or_projection, inverse=False):
    if not inverse:
        return _equisolid(angle_or_projection)
    return _equisolid_inverse(angle_or_projection)


@cfunc(float64(float64, bool_))
def equidistant(angle_or_projection, inverse=False):
    if not inverse:
        return _equidistant(angle_or_projection)
    return _equidistant_inverse(angle_or_projection)


@cfunc(float64(float64, bool_))
def orthographic(angle_or_projection, inverse=False):
    if not inverse:
        return _orthographic(angle_or_projection)
    return _orthographic_inverse(angle_or_projection)


@cfunc(float64(float64, bool_))
def stereographic(angle_or_projection, inverse=False):
    if not inverse:
        return _stereographic(angle_or_projection)
    return _stereographic_inverse(angle_or_projection)


@cfunc(float64(float64, bool_))
def thoby(angle_or_projection, inverse=False):
    if not inverse:
        return _thoby(angle_or_projection)
    return _thoby_inverse(angle_or_projection)




#  Copyright (c) 2022. Edson Moreira
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
#  documentation files (the "Software"), to deal in the Software without restriction, including without limitation
#  the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and
#  to permit persons to whom the Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all copies or substantial portions of
#  the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
#  BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
#  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

#
#  Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
#  documentation files (the "Software"), to deal in the Software without restriction, including without limitation
#  the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and
#  to permit persons to whom the Software is furnished to do so, subject to the following conditions:
#
#
from typing import Tuple

import numpy as np
from numba import njit, prange

from photonbend.core import SphereImage, LensImageType
from photonbend.utils import degrees_to_radians


@njit
def _projection_function(radius, standard_parallel, longitude_or_x: float, latitude_or_y: float, cardinal: bool):
    """ Gives the equirectangular projection function or it's inverse

    :param radius: The globe radius to project
    :param longitude_or_x:
    :param latitude_or_y:
    :param cardinal:
    :return:
    """
    x_or_long = _projection_x_longitude(radius, standard_parallel, longitude_or_x, cardinal)
    y_or_lat = _projection_y_latitude(radius, standard_parallel, latitude_or_y, cardinal)
    if not cardinal:
        x = x_or_long
        y = y_or_lat
        return x, y

    latitude = y_or_lat
    longitude = x_or_long
    return latitude, longitude


@njit
def _projection_x_longitude(radius: float, standard_parallel: float, longitude_or_x: float, cardinal: bool) -> float:
    if not cardinal:
        x = (radius * (longitude_or_x + np.pi)) * np.cos(standard_parallel)
        return x
    else:
        longitude = (longitude_or_x / (np.cos(standard_parallel) * radius)) - np.pi
        return longitude


@njit
def _projection_y_latitude(radius, standard_parallel: float, latitude_or_y: float, cardinal: bool) -> float:
    destiny_center_height = radius * np.pi / 2
    if not cardinal:
        y = destiny_center_height - (radius * latitude_or_y)
        return y
    else:
        latitude = (-(latitude_or_y - destiny_center_height)) / radius
        return latitude


@njit(parallel=False)
def make_projection(source: SphereImage, standard_parallel, desired_width) -> np.ndarray:
    # TODO Add a super sampler

    radius = desired_width / (np.pi * 2)

    destiny_width = int(np.abs(round(_projection_x_longitude(radius, standard_parallel, np.pi, False))))
    destiny_height = int(np.abs(round(_projection_y_latitude(radius, standard_parallel, -np.pi / 2, False))))
    destiny_array = np.zeros((destiny_height, destiny_width, 3), np.core.uint8)

    for row in prange(destiny_height):
        latitude = _projection_y_latitude(radius, standard_parallel, row, True)
        for column in prange(destiny_width):
            longitude = _projection_x_longitude(radius, standard_parallel, column, True)
            try:
                destiny_array[row, column, :] = source.get_from_spherical(latitude, longitude)
            except Exception:
                continue

    return destiny_array


@njit(parallel=False)
def make_sphere_image(source: np.array, lens, image_type: LensImageType, fov: float,
                      rotation: Tuple[float, float, float]):
    source_height, source_width = source.shape[:2]
    standard_parallel = np.arccos(min(1.0, source_width / 2 / source_height))

    # -0.1 is a dirty trick to go around rounding problems and continuous nature of geometry
    radius = (source_height - 0.1) / np.pi

    destiny_height = int(round(np.pi * radius))
    destiny_width = destiny_height

    if image_type == LensImageType.DOUBLE_INSCRIBED:
        destiny_width = destiny_width * 2

    d_sphere = SphereImage(np.zeros((destiny_height, destiny_width, 3), np.core.uint8), image_type,
                           degrees_to_radians(fov), lens)

    if rotation != (0, 0, 0):
        rotation_rad = list(map(degrees_to_radians, rotation))
        pitch, yaw, roll = rotation_rad
        d_sphere.set_rotation(pitch, roll, yaw)

    for row in range(destiny_height):
        for column in range(destiny_width):
            try:
                latitude, longitude = d_sphere.translate_cartesian_to_spherical(column, row)

                x, y = _projection_function(radius, standard_parallel, longitude, latitude, False)

                x = int(round(x))
                y = int(round(y))

                if x < 0 or x >= source_width:
                    continue
                if y < 0 or y >= source_height:
                    continue
                value = source[y, x]
                d_sphere.set_to_cartesian(column, row, value)
            except Exception:
                continue
    return d_sphere


def compute_best_width(source: SphereImage) -> int:
    result = 2 * (int(round(2 * source.lens_image.shape[0])) // 2)
    return result



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['photonbend',
 'photonbend.core',
 'photonbend.exceptions',
 'photonbend.lens',
 'photonbend.projections',
 'photonbend.scripts',
 'photonbend.scripts.commands',
 'photonbend.utils']

package_data = \
{'': ['*']}

install_requires = \
['Pillow>=9.0.1,<10.0.0',
 'click>=8.0.4,<9.0.0',
 'numba>=0.55.1,<0.56.0',
 'numpy>=1.18,<1.22',
 'scipy>=1.8.0,<2.0.0']

entry_points = \
{'console_scripts': ['photonbend = photonbend.scripts.main:main']}

setup_kwargs = {
    'name': 'photonbend',
    'version': '0.1.0',
    'description': 'photonbend deals with camera pictures (especially fisheye) and some projections like equirectangular',
    'long_description': None,
    'author': 'Edson Moreira',
    'author_email': 'w.moreirae@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



#  Copyright (c) 2022. Edson Moreira
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
#  documentation files (the "Software"), to deal in the Software without restriction, including without limitation
#  the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and
#  to permit persons to whom the Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all copies or substantial portions
#  of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
#  BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
#  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
from typing import Tuple, Callable

import numpy as np
from numba import njit


@njit
def decompose(a_complex_number):
    """Decomposes a complex number into it's real and imaginary parts and returns them as integers.
    To turn the parts into integers, it rounds them first and the proceed s to cast them.

    :param a_complex_number: any complex number
    :return: a tuple of integers representing the real and imaginary parts of the original number
    """
    x = int(round(a_complex_number.real))
    y = int(round(a_complex_number.imag))
    return x, y


@njit
def angle_from_vector(vector: complex):
    u_vector = unit_vector(vector)
    theta = np.log(vector).imag
    return theta


@njit
def vector_magnitude(vector: complex):
    """Computes the magnitude of a complex number vector

        :param vector: A complex number
        :return: A floating point scalar of the vector magnitude
        """
    return np.sqrt(vector.real ** 2 + vector.imag ** 2)


@njit
def degrees_to_radians(degrees: float):
    return degrees / 180 * np.pi


@njit
def radians_to_degrees(rad: float) -> float:
    return rad / np.pi * 180.0


@njit
def dpi_to_dpmm(dpi: int) -> float:
    mm_in_a_inch = 25.4
    return dpi / mm_in_a_inch


@njit
def vector_from_coordinates(x, y) -> complex:
    vector = complex(x, y)
    return vector


@njit
def unit_vector(vector: complex) -> complex:
    magnitude = vector_magnitude(vector)
    if 0 == magnitude:
        return complex(0, 0)
    u_vector = vector / magnitude
    return u_vector


@njit
def c_round(complex_number) -> Tuple[int, int]:
    return int(round(complex_number.real)), int(round(complex_number.imag))


@njit
def vector_to_focal_units(vector: complex, focal_distance: float, pixels_per_f_distance: float) -> float:
    """Calculates the projection of a pixel.

    With the projection of a pixel, you can use an inverse_mapping_function to get the lens angle of it
    The projection is 'normalized' using the focal distance as the factor, therefore, 'normalized' may be a
    misnomer in case of a  any mapping function that produces values bigger than 1.
    """
    opposite_side = vector_magnitude(vector)
    adjacent_side = focal_distance * pixels_per_f_distance

    normalized_projection_magnitude = np.absolute(opposite_side / adjacent_side)

    return normalized_projection_magnitude


@njit
def calculate_lens_angle(vector: complex, f_distance: float, dpi: float,
                         mapping_function: Callable[[float, bool], float]):
    normalized_magnitude = vector_to_focal_units(vector, f_distance, dpi)
    angle = mapping_function(normalized_magnitude, True)
    return angle * 2


@njit
def calculate_pixels_per_f_distance(vector: complex, angle: float, f_distance: float,
                                    mapping_function: Callable[[float, bool], float]):
    half_angle = angle / 2
    quasi_magnitude = mapping_function(half_angle, False) * f_distance
    v_magnitude = vector_magnitude(vector)
    pixels_pfd = v_magnitude / quasi_magnitude
    return pixels_pfd


# TODO MAKE A TRANSLATE TO GEODESIC

@njit
def _Z_get_360_longitude(longitude):
    new_longitude = longitude % FULL_CIRCLE
    return new_longitude


@njit
def _Z_get_180_longitude(longitude):
    new_longitude = longitude % FULL_CIRCLE
    if np.pi < new_longitude:
        new_longitude -= (2 * np.pi)
    return new_longitude


@njit
def decompose(a_complex_number):
    """Decomposes a complex number into it's real and imaginary parts and returns them as integers.
    To turn the parts into integers, it rounds them first and the proceed s to cast them.

    :param a_complex_number: any complex number
    :return: a tuple of integers representing the real and imaginary parts of the original number
    """
    x = int(round(a_complex_number.real))
    y = int(round(a_complex_number.imag))
    return x, y


@njit
def weighted_sum(v1, v2, w1, w2):
    sum_weight = w1 + w2
    v1_arr = np.zeros(3, np.core.uint64)
    v2_arr = np.zeros(3, np.core.uint64)

    v1_arr[:] = v1[:]
    v2_arr[:] = v2[:]

    v1 = v1 * w1
    v2 = v2 * w2
    v3 = v1 + v2
    v3 = v3 / sum_weight
    v3 = v3.astype(np.core.uint8)
    return v3


@njit
def _2ints(v1, v2):
    return int(v1), int(v2)
